// Mega644 version

#include<inttypes.h>
#include<avr/io.h>
#include<avr/interrupt.h>
#include<stdio.h>
#include<avr/pgmspace.h>
#include<string.h>


//set up the debugging utility ASSERT
//#define __ASSERT_USE_STDERR
//#include <assert.h>

// serial communication library
#include "uart.h"

//defining waiting times
#define SIGNDIFFERENCE 400
#define SYMBOLWAITTIME 200
#define COUNTWAITTIME 50

//defining value of symbols
#define ROCK 0
#define PAPER 3
#define SCISSORS 1

//defining how many hand hits
#define FINALHITCOUNT 3

void initialize(void); //all the usual mcu stuff 

int8_t lcd_buffer[25];	// LCD display buffer

//counting hand hits
volatile char p1Counter;
volatile char p2Counter;

//counter variables for timer to prevent multiple triggerings from accel
volatile char p1CountTimer;
volatile char p2CountTimer;

//indicates if game is in progress
volatile char gameStarted;

//indicates hand hit count is even
volatile char evenCount;

//trigger printing
volatile char p1Print;
volatile char p2Print;

//measures time differences between displayed symbols
volatile char cheatFlag;
volatile unsigned int diffTime;

//wait time for symbol
volatile unsigned int waitTime;

//total wins and current winner
int p1Wins;
int p2Wins;
char gameWinner;

//displayed symbols by each player
char p1Symbol;
char p2Symbol;

//subfunction stubs
char determineWinner(char p1, char p2);
void resetGame();

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
         
//***********************************************************/
//Player 1 accelerometer count
//External interrupt on port D.2
ISR (INT0_vect)
{   
	//only count up to max number of hand hits, don't allow two hits to occur too close together
	if(p1Counter<=FINALHITCOUNT-1 && p1CountTimer==0)
	{
		p1CountTimer = COUNTWAITTIME;

		//only allow p1 to get 1 hand hit in front of p2
		if(p1Counter<=p2Counter)
			p1Counter++;
		
		//on first hand hit, start game
		if(p1Counter == 1)
			gameStarted = 1;
		
		//if count is even, stop timeout counter
		if(p1Counter == p2Counter)
		{
			evenCount = 1;
			diffTime = SIGNDIFFERENCE;
		}
		else
			evenCount = 0;
		
		//set count printer
		p1Print = 1;
	}
}  
//END Player 1 External ISR
//***********************************************************/


//***********************************************************/
//Player 2 accelerometer count
//External interrupt on port B.2
ISR (INT2_vect)
{   

	//only count up to max number of hand hits, don't allow two hits to occur too close together
	if(p2Counter<=FINALHITCOUNT-1 && p2CountTimer==0)
	{
		p2CountTimer = COUNTWAITTIME;

		//only allow p2 to get 1 hand hit in front of p1
		if(p2Counter<=p1Counter)
			p2Counter++;
		
		//on first hand hit, start game
		if(p2Counter == 1)
			gameStarted = 1;
		
		//if count is even, stop timeout counter
		if(p1Counter == p2Counter)
		{
			evenCount = 1;
			diffTime = SIGNDIFFERENCE;
		}
		else
			evenCount = 0;
		
		//set count printer
		p2Print = 1;
	}
}  
//END Player 2 External ISR
//***********************************************************/


//***********************************************************/
//Timeout Detection
//timer 0 overflow ISR
//millisecond counter
ISR (TIMER0_COMPA_vect) 
{   
	//player 1 hit timer - to prevent double triggerings on hits
	if(p1CountTimer>0)
		p1CountTimer--;

	//player 2 hit timer - to prevent double triggerings on hits
	if(p2CountTimer>0)
		p2CountTimer--;

	//if game is started and count is not even, start timeout
	if(gameStarted == 1 && evenCount == 0)
	{
		if(diffTime>0)	diffTime--;
		
		//if timeout, throw cheat flag to end game
		if(diffTime==0)
		{
			cheatFlag = 1;
		}	
	}
	//if symbols are displayed, wait for flex sensors to stabilize
	else if(p1Counter==FINALHITCOUNT && p2Counter==FINALHITCOUNT)
	{
		if(waitTime>0)
		{
			waitTime--;
			gameStarted = 0;
		}
	}
}
//END Timeout Detection ISR
//***********************************************************/


//***********************************************************/       
//Main Program Loop
//Entry point and task scheduler loop
int main(void)
{  
	initialize();
  
  	//main task scheduler loop 
  	while(1)
  	{ 
		//player 1 hand hit counter printer
		if(p1Print==1)
		{
			fprintf(stdout, "Player 1 Count: %d\n",p1Counter);
			p1Print = 0;
		}
		
		//player 2 hand hit counter printer
		if(p2Print==1)
		{
			fprintf(stdout, "Player 2 Count: %d\n",p2Counter);
			p2Print = 0;
		}
		
		//if cheat flag has been thrown, end game
		if(cheatFlag==1)
		{
			fprintf(stdout, "Timeout: Game Ended\n\n");
			resetGame();
			cheatFlag = 0;
		}
		//when symbols are displayed, find the winner
		else if(p1Counter==FINALHITCOUNT && p2Counter==FINALHITCOUNT && waitTime==0)
		{
			//Read Player 1 symbol
			//index finger on D.3
			//ring finger on D.4
			p1Symbol = (PIND>>3) & 0x03;
			fprintf(stdout,"p1Symbol: %d\n",p1Symbol);
			
			//Read Player 2 symbol
			//index finger on B.3
			//ring finger on B.4
			p2Symbol = (PINB>>3) & 0x03;
			fprintf(stdout,"p2Symbol: %d\n",p2Symbol);
			
			//determine winner
			gameWinner = determineWinner(p1Symbol, p2Symbol);
			
			//print winner and increment win counters
			//1 = p1 wins
			//2 = p2 wins
			//3 = tie
			//0 = error in detection
			switch(gameWinner)
			{
				case 0: 
					fprintf(stdout,"Game error!\n\n");
					break;
				case 1:
					fprintf(stdout,"Player 1 wins\n\n");
					p1Wins++;
					break;
				case 2:
					fprintf(stdout,"Player 2 wins\n\n");
					p2Wins++;
					break;
				case 3:
					fprintf(stdout,"Tie Game\n\n");
					break;
				default:
					fprintf(stdout,"Error\n\n");
			}
			
			//print total wins amounts
			fprintf(stdout, "Player 1 Total Wins: %d\n",p1Wins);
			fprintf(stdout, "Player 2 Total Wins: %d\n\n",p2Wins);

			//reset game
			resetGame();
		}			
  	}
}
//END Main
//***********************************************************/


//***********************************************************/
//Reset Game
//Resets all counters and game states
void inline resetGame(void)
{
	diffTime = SIGNDIFFERENCE;
	waitTime = SYMBOLWAITTIME;
	p1CountTimer = 0;
	p2CountTimer = 0;
	p1Counter = 0;
	p2Counter = 0;
	gameStarted = 0;
	evenCount = 1;
	p1Print = 0;
	p2Print = 0;
	p1Symbol = 0;
	p2Symbol = 0;
	gameWinner = 0;
}
//END Reset Game
//***********************************************************/


//***********************************************************/
//Determine Winner
//Takes in the displayed symbols by each player and determines winner
//Returned Values: 1 (p1 wins), 2 (p2 wins), 3 (tie), 0 (error)
//Also prints out each player's symbol
char determineWinner(char p1, char p2)
{
	char winner;
	//covers all possibilities for player 1 displaying Rock
	if(p1==ROCK)
	{
		fprintf(stdout,"P1: ROCK\n");
		if(p2==ROCK)
		{
			fprintf(stdout,"P2: ROCK\n");
			winner = 3;
		}
		else if(p2 == PAPER)
		{
			fprintf(stdout,"P2: PAPER\n");
			winner = 2;
		}
		else if(p2 == SCISSORS)
		{
			fprintf(stdout,"P2: SCISSORS\n");
			winner = 1;
		}
		//player 2 symbol error
		else
			winner = 0;
	}
	//covers all possibilities for player 1 displaying Paper
	else if(p1 == PAPER)
	{
		fprintf(stdout,"P1: PAPER\n");
		if(p2==ROCK)
		{
			fprintf(stdout,"P2: ROCK\n");
			winner = 1;
		}
		else if(p2 == PAPER)
		{
			fprintf(stdout,"P2: PAPER\n");
			winner = 3;
		}
		else if(p2 == SCISSORS)
		{
			fprintf(stdout,"P2: SCISSORS\n");
			winner = 2;
		}
		//player 2 symbol error
		else
			winner = 0;
	}
	//covers all possibilities for player 1 displayer Scissors
	else if(p1 == SCISSORS)
	{
		fprintf(stdout,"P1: SCISSORS\n");
		if(p2==ROCK)
		{
			fprintf(stdout,"P2: ROCK\n");
			winner = 2;
		}
		else if(p2 == PAPER)
		{
			fprintf(stdout,"P2: PAPER\n");
			winner = 1;
		}
		else if(p2 == SCISSORS)
		{
			fprintf(stdout,"P2: SCISSORS\n");
			winner = 3;
		}
		//player 2 symbol error
		else
			winner = 0;
	}
	//player 1 symbol error
	else
		winner = 0;
		
	//return final winner decision
	return winner;
}  
//END Determine Winner
//***********************************************************/



//***********************************************************/
//Initialize
//Set it all up
void initialize(void)
{	
	//initialize all game states and variables
	resetGame();
	cheatFlag = 0;
	p1Wins = 0;
	p2Wins = 0;

	//set ports D and B to inputs
	DDRB = 0x00;
	DDRD = 0x00;
	
	//set up external interrupts
	EIMSK = 1<<INT0 | 1<<INT2 ; // turn on int0 and int2
	EICRA = 0b00110011 ;       // rising edge for both int0 and int2


	//set up timer 0 for 1 mSec timebase 
	TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR 
	OCR0A = 249;  		//set the compare re to 250 time ticks
	//set prescalar to divide by 64 
	TCCR0B= 3; //0b00001011;	
	// turn on clear-on-match
	TCCR0A= (1<<WGM01) ;
	

	//init the UART -- uart_init() is in uart.c
  	uart_init();
  	stdout = stdin = stderr = &uart_str;
  	fprintf(stdout,"Starting...\n\r");
      
	//crank up the ISRs
	sei();
}  
//END Initialize
//***********************************************************/
